
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/fb.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

// -----------------------------------------------------------------------------
static struct fb_var_screeninfo vinfo;
static struct fb_fix_screeninfo finfo;
static uint16_t * fb_ptr = NULL;
static size_t fb_size = 0;

// -----------------------------------------------------------------------------
static void draw_rect(uint32_t xpos, uint32_t ypos,
					  uint32_t lenght, uint32_t height,
					  uint16_t color)
{
	int i, j;

	for ( i = ypos; i < ypos + height; i++ )
	{
		for ( j = xpos; j < xpos + lenght; j++ )
		{
			fb_ptr[j + i * vinfo.xres] = color;
		}
	}
}

// -----------------------------------------------------------------------------
static void blinking_rect(uint32_t period)
{
	while (1)
	{
		struct timespec delay = { .tv_sec = 0, .tv_nsec = period * 1000 * 1000 };

		// zeleny + smazani
		draw_rect(400, 240, 10, 10, 0x001F);
		nanosleep(&delay, NULL);
		draw_rect(400, 240, 10, 10, 0x0000);

		// cerveny + smazani
		draw_rect(420, 240, 10, 10, 0xf800);
		nanosleep(&delay, NULL);
		draw_rect(420, 240, 10, 10, 0x0000);
	}
}

// -----------------------------------------------------------------------------
int main (int argc, char ** argv)
{
	int res = 0;
	int fbFile = 0;
	void * mmem_ptr = NULL;
	size_t mmem_size = 0;

	fbFile = open("/dev/fb0", O_RDWR);
	if ( fbFile == 0 )
	{
		printf("Error: cannot open fb device!\n");
		exit(1);
	}
	
	printf("device //dev//fb0 opened.\n");

	res = ioctl(fbFile, FBIOGET_FSCREENINFO, &finfo);
	if ( res < 0 )
	{
		printf("Error: reading fixed info failed!\n");
		exit(2);
	}

	res = ioctl(fbFile, FBIOGET_VSCREENINFO, &vinfo);
	if ( res < 0 )
	{
		printf("Error: reading variable info failed!\n");
		exit(3);
	}

	mmem_size = (vinfo.xres * vinfo.yres * vinfo.bits_per_pixel) / 8;

	mmem_ptr = mmap(NULL, mmem_size, PROT_READ | PROT_WRITE, MAP_SHARED, fbFile, 0);
	if ( mmem_ptr == MAP_FAILED )
	{
		printf("Error(%d): mmaping fb device failed!\n", errno);
		perror("Chyba");
		exit(4);
	}

	printf("fb mmapped, memsize = %uB\n", mmem_size);

	// ukazatek na mmpaed pamet prevedeme na ukazatel na framebuffer - co pozice
	// to pixel s danou barevnou hloubkou, ted 16bitovou
	fb_ptr = (uint16_t *) mmem_ptr;
	fb_size = mmem_size / sizeof(uint16_t);

	// nakresli 5 x 5px od pocatku displeje modry ctverec 10 x 10 pixelu
	draw_rect(5, 5, 10, 10, 0x001F);

	// nakresli fialovy ctverec 15x15 px na pozici 20x20 px od zacatku
	// bez aplikace patche by se posledni pixel ctverce nezobrazil, protoze se
	// nachazi na liche pozici. Pritom do radice SSD1963 je zapsan.
	draw_rect(20, 20, 15, 15, 0xF81f);

	// blikajici ctverce, perioda v milisec
	blinking_rect(200);

	printf("done\n");

	munmap(mmem_ptr, mmem_size);
	close(fbFile);
	
	return EXIT_SUCCESS;
}
